% PROCESS Run image enhancement process
%
%   Available options:
%
%     mono [$]    - Convert image to a grayscale image. 
%       hsv         Use the V value of HSV color-space.
%       mean        Use the mean color between R, G and B.
%       ycbcr       Use the Y value of YCBCR color-space (default).
%
%     hist $      - Equalize the histogram for a Gaussian distribution.
%                   Default value is 1.5.
%     nohist      - Don't use histogram equalization. The second argument for
%                   the procedure may be an already histogram-equalized image.
%
%     color $     - Color scheme for RETINEX calculation.
%       hsv         Use the V value of HSV color-space.
%       rgb         Use the R, G and B channels separately (default).
%       ycbcr       Use the Y value of YCBCR color-space.
%     colcorr [$] - Use color correction scheme [NASA MSR/CR].
%       linear      Use ID (no transform) in CR.
%       log         Use LOG2 in CR (default).
%     gamma $     - Set Gamma correction value.
%     retinex $$  - Set RETINEX alpha and beta parameters. 
%     skipretinex - Skip retinex process. The second and third parameters should 
%                   be the two retinex images.
%
%     fixhist [$] - Fix the combined image histogram with the inverse of Gaussian
%                   distribution. Default value is 1.0 or 2/3 of hist.
%     fixhists $  - Fix histogram for given sigma.
%
%     stretch [$] - Image stretching. Default is 0.5% of the pixels.
%       all [$]     Stretch all channels together (default).
%       sep [$]     Stretch all channels separately.
%
%     show        - Show the final result. Turned on if there is no output.
%     showall     - Show all the intermediate images.
%     save $      - Write the final image to disk.
%     saveall $   - Write all the images to disk.
%       <fn>        Base filename without extension (PNG).
%     title $     - Give a title for all the images, applicable for 'show' 
%                   or 'showall'.

function [L, S,J,K] = process(I, varargin);

skipping = 0; 
mono = 0; colorscheme = 1;
colorcorr = 0;
showtemp = 0; showres = 0;
save = 0; saveres = 0;
histsigma = 1.5;
titlestr = '';
fixhist = 0;
alpha = 0.0001; beta = 0.1; gamma = inf;
stretch = 0;

% Parse input arguments
skip = 0;
for i=1:length(varargin),
  if skip,
    skip = skip-1;
  elseif isstr(varargin{i}),
    switch varargin{i}
    case 'colcorr'
      colorcorr = 2;
      if i<length(varargin), 
        switch varargin{i+1}
        case 'linear'
          colorcorr = 1;
          skip = 1;
        case 'log'
          colorcorr = 2;
          skip = 1;
        end;
      end;
    case 'color'
      switch varargin{i+1}
      case 'hsv'
        colorscheme = 3;
      case 'rgb'
        colorscheme = 1;
      case 'ycbcr'
        colorscheme = 2;
      end;
      skip = 1;
    case 'fixhist'
      if i<length(varargin) & isnumeric(varargin{i+1}), 
        fixhist = varargin{i+1};
        skip = 1;
      else
        fixhist = histsigma*2/3;
      end;
    case 'fixhists'
      fixhistsigma = varargin{i+1};
      fixhist = -1;
      skip = 1;
    case 'gamma'
      gamma = varargin{i+1};
      skip = 1;
    case 'hist'
      histsigma = varargin{i+1};
      skip = 1;
    case 'mono'
      mono = 2;
      colorscheme = 0;
      if i<length(varargin), 
        switch varargin{i+1}
        case 'hsv'
          mono = 3;
          skip = 1;
        case 'mean'
          mono = 1;
          skip = 1;
        case 'ycbcr'
          mono = 2;
          skip = 1;
        end;
      end;
    case 'nohist'
      skipping = 1;
      S = varargin{1};
      if ~isnumeric(S), S = I; end;
    case 'retinex'
      alpha = varargin{i+1};
      beta = varargin{i+2};
      skip = 2;
      if i+3<=length(varargin) & isnumeric(varargin{i+3}), 
        gamma = varargin{i+3};
        skip = 3;
      end;
    case 'save'
      saveres = 1;
      name = varargin{i+1};
      skip = 1;
    case 'saveall'
      save = 1;
      saveres = 1;
      name = varargin{i+1};
      skip = 1;
    case 'show'
      showres = 1;
    case 'showall'
      showtemp = 1;
      showres = 1;
    case 'skipretinex'
      J = varargin{1};
      K = varargin{2};
      skipping = 2;
    case 'stretch'
      stretch = 1;
      if i<length(varargin) & isstr(varargin{i+1}),
        switch varargin{i+1}
        case 'all'
          stretch = 1;
          skip = 1;
        case 'sep'
          stretch = 2;
          skip = 1;
        end;
      end;
      if i+skip<length(varargin) & isnumeric(varargin{i+skip+1}),
        skip = skip+1;
        stramount = varargin{i+skip}/100;
      else
        stramount = 0.005;
      end;
    case 'title'
      titlestr = varargin{i+1};
      skip = 1;

    otherwise
      disp(['Warning: Unknown switch ''', varargin{i}, '''.']);
    end;
  end;
end;

% If no output, draw the result on the screen.
if nargout==0 & ~saveres,
    showres = 1;
end;


% Load the image if given filename
if isstr(I),
  fname = I;
  [p, n, e] = fileparts(fname);
  disp(sprintf('Reading image from %s.',fname));
  if (strcmp(e,'.ppm16')),
    I = readppm16(fname);
  else
    I = im2double(imread(fname));
  end;
end;

% Create output directory, if output directory does not exist
if saveres,
    pth = fileparts(name);
    if ~isempty(pth) & ~exist(pth, 'dir'),
        mkdir(pth);
    end;
end;

% If monochromatic, convert image
if size(I,3)==1,
  colorscheme = 0; % Force GS colorscheme
elseif mono, 
  disp('Converting image to monochrome.');
  I = convertToMono(I,mono);
  if skipping & size(S,3)==3, % in the case of 'nohist'
    S = convertToMono(S,mono);
  end;
end;

% Smooth image
if ~skipping,
  disp('Smoothing the image.');
  for n=1:size(I,3),
    I(:,:,n) = eFilter2(fspecial('gaussian',5,0.4),I(:,:,n));
  end;
  if showtemp, figure; imshow(I); title(titlestr); end;
  if save, imwrite(I,sprintf('%s.org.png',name)); end;
end;

% Equalize the image histogram
if ~skipping,
  disp(sprintf('Histogram equalization [sigma=%f].',histsigma));
  if size(I,3)==3, % Overcome a constant color channel.
    M = rgb2y(I);
    for n=1:size(I,3),
      I(:,:,n) = I(:,:,n)*0.99 + M*0.01;
    end;
    clear M;
  end;
  S = fullHisteq(I,histsigma);
  if showtemp, figure; imshow(S); title(titlestr); end;
  if save, imwrite(S,sprintf('%s.hist.png',name)); end;
end;

% Calculate the two RETINEXes
if skipping<2, 
  switch colorscheme
  case 0 % Grayscale channel
    disp('Calculating RETINEX.');
    J = retinex2(S, alpha, beta, gamma); % regular Retinex
    K = 1-retinex2(1-S, alpha, beta, gamma); % inverted Retinex
  
  case 1 % RGB channels
    disp('Calculating RETINEX on RGB.');
    J(:,:,1) = retinex2(S(:,:,1), alpha, beta, gamma);
    J(:,:,2) = retinex2(S(:,:,2), alpha, beta, gamma);
    J(:,:,3) = retinex2(S(:,:,3), alpha, beta, gamma);
    K(:,:,1) = 1-retinex2(1-S(:,:,1), alpha, beta, gamma);
    K(:,:,2) = 1-retinex2(1-S(:,:,2), alpha, beta, gamma);
    K(:,:,3) = 1-retinex2(1-S(:,:,3), alpha, beta, gamma);
    
  case 2 % YCBCR
    disp('Calculating RETINEX on Y of YCBCR.');
    J = rgb2ycbcr(S);
    K = J;
    J(:,:,1) = retinex2(J(:,:,1), alpha, beta, gamma);
    K(:,:,1) = 1-retinex2(1-K(:,:,1), alpha, beta, gamma); % 'rgb2ycbcr' is a linear function
    J = ycbcr2rgb(J);
    K = ycbcr2rgb(K);
    
  case 3 % HSV
    disp('Calculating RETINEX on V of HSV.');
    J = rgb2hsv(S);
    K = rgb2hsv(1-S);
    J(:,:,3) = retinex2(J(:,:,3), alpha, beta, gamma);
    K(:,:,3) = retinex2(K(:,:,3), alpha, beta, gamma);
    J = hsv2rgb(J);
    K = 1-hsv2rgb(K);
  end;
  
  if showtemp, figure; imshow(truncate(J)); title(titlestr); end;
  if save, imwrite(truncate(J),sprintf('%s.rtx1.png',name)); end;
  if showtemp, figure; imshow(truncate(K)); title(titlestr); end;
  if save, imwrite(truncate(K),sprintf('%s.rtx2.png',name)); end;
end;

  
% Combine the two RETINEXes
disp('Combining RETINEXes.');
L = (J+K)*0.5;
if showtemp, figure; imshow(truncate(L)); title(titlestr); end;
if save, imwrite(truncate(L),sprintf('%s.comb.png',name)); end;
if nargout<3,
  clear J K
end;


% Fix the histogram
if fixhist==-1,
  fixhist = fixhistsigma/std(L(:));
  if (fixhist<1), fixhist=0; end;
end;
if fixhist>0
  disp(sprintf('Fixing image histogram [sigma=%f].', fixhist));
  % Overcome memory problems with large images...
  % L = erf((L-0.5)*2*fixhist)/(2*erf(fixhist)) + 0.5;
  for i=1:1000000:prod(size(L)),
    j = min(i+1000000-1,prod(size(L)));
    L(i:j) = erf((L(i:j)-0.5)*2*fixhist)/(2*erf(fixhist)) + 0.5;
  end;
  if showtemp, figure; imshow(truncate(L)); title(titlestr); end;
  if save, imwrite(truncate(L),sprintf('%s.fix.png',name)); end;
end;

% Color Correction
if colorcorr & size(L,3)==3,
% B = 1./rgb2y(I);
  B = mean(I,3);
  nz = find(B);
  B(nz) = 1./B(nz);
  if colorcorr==1,
    disp('Color Resotration (linear).');
    L(:,:,1) = L(:,:,1) .* (I(:,:,1).*B);
    L(:,:,2) = L(:,:,2) .* (I(:,:,2).*B);
    L(:,:,3) = L(:,:,3) .* (I(:,:,3).*B);
  else
    disp('Color Resotration (logarithmic).');
    L(:,:,1) = L(:,:,1) .* log2(1+I(:,:,1).*B);
    L(:,:,2) = L(:,:,2) .* log2(1+I(:,:,2).*B);
    L(:,:,3) = L(:,:,3) .* log2(1+I(:,:,3).*B);
  end;
  clear B
  if showtemp, figure; imshow(truncate(L)); title(titlestr); end;
  if save, imwrite(truncate(L),sprintf('%s.cc.png',name)); end;
end;

% Note: L is NOT truncated (yet)!!!

% Image Stretching
if stretch,
  if stretch==1 | mono, % Stretch all the channels togather
    disp(sprintf('Image Stretching by %2.1f%%.',stramount*100));
    L = stretchImage(L,stramount);
  else % Stretch each channel separately
    disp(sprintf('Image Stretching (separately) by %2.1f%%.',stramount*100));
    for n=1:size(L,3),
      L(:,:,n) = stretchImage(L(:,:,n),stramount);
    end;
  end;
  if showtemp, figure; imshow(truncate(L)); title(titlestr); end;
  if save, imwrite(truncate(L),sprintf('%s.stretch.png',name)); end;
end;

if showres, figure; imshow(truncate(L)); title(titlestr); end;
if saveres, 
    disp(sprintf('Saving image to %s.png.',name));
    imwrite(truncate(L),sprintf('%s.png',name)); 
end;
disp(' ');


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function M = convertToMono(I,method);

switch method,
case 1
  M = mean(I,3);
case 2
  M = rgb2y(I);
case 3
  M = rgb2hsv(I);
  M = I(:,:,3);
end;
